/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <linux/dprobes.h>
#include <linux/dprobes_in.h>

#include "main.h"
#include "ppdf.h"

#define MAX_TYPES	256
#define MAX_GROUPS	256

char * cmdline = NULL;
char * cmdline_ptr = NULL;
char * cmdline_end = NULL;
int cmdline_len = 0;

char * rpnfile = NULL;
char * ppdffile = NULL;
static char * prepfile = NULL;
char * symfile = NULL;
char * modname = NULL;
static unsigned short index_from = 0;
static unsigned short index_to = 0;

unsigned int rpnlineno = 1;

unsigned long dp_cmd = 0;
static int group_count = 0;
static int type_count = 0;
static int g_index[MAX_GROUPS], t_index[MAX_TYPES];
static label_table_t cmdl_group_list = {"cmdl group list", MAX_GROUPS, 0, NULL};
static label_table_t cmdl_type_list = {"cmdl type list", MAX_TYPES, 0, NULL};

extern list_entry_t * lookup(label_table_t * list, const char * name);
extern void add_to_list(label_table_t *list, char *name, unsigned short off);
extern int lookup_index(label_table_t * list, const char * name);
extern int group_index(char * name);
extern int type_index(char * name);

void add_rpn_group(char * name);
void add_rpn_type(char * name);

extern void set_range(unsigned long from, unsigned long to);

extern void display_help(void);

struct defines {
	char * key;
	char * val;
};

#define MAX_DEFINES 32
struct defines * defines_list = NULL;
int num_defines = 0;
static int max_defines = MAX_DEFINES;

void cmdlerror (const char * err)
{
	printf("Syntax error: %s\n", err);
	printf("Try `dprobes --help' for more information\n");
	exit(1);
}

void add_rpn_group(char * name)
{
	char buff[32];
	list_entry_t * dup;

	if (strlen(name) > 15) {
		printf("Warning: group %s truncated to 15 characters.\n", name);
		name[15] = '\0';
	}

	dup = lookup(&cmdl_group_list, name);
	if (dup) {
		sprintf(buff, "group %s is duplicate.", name);
		cmdlerror(buff);
	}
	++group_count;
	cmdl_group_list.entry = realloc(cmdl_group_list.entry,
				group_count * sizeof(list_entry_t));
	if (!cmdl_group_list.entry) {
		fatal_error("realloc failed\n");
	}
	add_to_list(&cmdl_group_list, name, 0);
}

void add_rpn_type(char * name)
{
	char buff[32];
	list_entry_t * dup;

	if (strlen(name) > 15) {
		printf("Warning: type %s truncated to 15 characters.\n", name);
		name[15] = '\0';
	}

	dup = lookup(&cmdl_type_list, name);
	if (dup) {
		sprintf(buff, "type %s is duplicate.", name);
		cmdlerror(buff);
	}
	++type_count;
	cmdl_type_list.entry = realloc(cmdl_type_list.entry,
				type_count * sizeof(list_entry_t));
	if (!cmdl_type_list.entry) {
		fatal_error("realloc failed\n");
	}
	add_to_list(&cmdl_type_list, name, 0);
}

char * strip_quotes(char * str)
{
	char * in_str = str;
	char * out_str;
	int len;
	char last_ch;
	if (*in_str == '\"')
		in_str++;
	while(*in_str == ' ')
		in_str++;
	len = strlen(in_str);
	last_ch = in_str[len-1];
	if (last_ch == '\"' || last_ch == '\r' || last_ch == '\n')
		len--;
	while(in_str[len-1] == ' ')
		len--;
	out_str = (char *) malloc(len+1);
	if (!out_str)
		fatal_error("malloc failed");
	memcpy(out_str, in_str, len);
	out_str[len] = '\0';
	return out_str;
}

int fatal_error(char * str)
{
	printf("dprobes: fatal error: %s\n", str);
	exit(1);
}

void no_options()
{
	printf("dprobes: missing/incorrect arguments\n");
	printf("Try `dprobes --help' for more information\n");
	exit(1);
}

void set_cmd(unsigned long opt)
{
	if (dp_cmd & opt) {
		cmdlerror("option specified multiple times\n");
	} else {
		dp_cmd |= opt;
	}
}

void set_string_val(char **dest, char * src)
{
	if (*dest) {
		cmdlerror("option specified multiple times\n");
	} else {
		*dest = src;
	}
}

void set_merge_replace(unsigned long opt)
{
	if (dp_cmd & DP_MERGE || dp_cmd & DP_REPLACE) {
		cmdlerror("conflicting options or options specified multiple times\n");
	} else {
		dp_cmd |= opt;
	}
}

void add_define(char * key, char * val)
{
	if (num_defines >= max_defines) {
		max_defines += 32;
		defines_list = (struct defines *)realloc((void *)defines_list,
			(max_defines * sizeof(*defines_list)));
		if (!defines_list) {
			fatal_error("realloc failed\n");
		}
	}
	defines_list[num_defines].key = key;
	defines_list[num_defines].val = val;
	num_defines++;
}

void add_define_num(char * key, unsigned long ulval)
{
	char val[20] = "\0";
	sprintf(val, "0x%lx", ulval);
	add_define(key, strdup(val));
}

void set_range(unsigned long from, unsigned long to)
{
	if (from > to) {
		index_from = (unsigned short)to;
		index_to = (unsigned short)from;
	} else {
		index_from = (unsigned short)from;
		index_to = (unsigned short)to;
	}
}

extern int dp_compile(char *rpnfile,
		char *symfile,
		struct dp_pgm_struct *dp_pgm);

void display_version()
{
	printf("IBM Dynamic Probes %d.%d.%d\n",
		DP_MAJOR_VER,DP_MINOR_VER,DP_PATCH_VER);
	printf("Copyright (c) 2000 - IBM Corporation\n");
}

#define DPROBES_MODULE 1
#ifdef DPROBES_MODULE

int dp_fd;

static int dprobes(unsigned long cmd, void * req, void * result)
{
	struct dp_ioctl_arg dp_arg;
	dp_arg.cmd = cmd;
	dp_arg.input = req;
	dp_arg.output = result;

	return ioctl(dp_fd, 0, (unsigned long)&dp_arg);	
}

static int dp_device_open(void)
{
	dp_fd = open("/dev/dprobes", O_RDWR);
	if (dp_fd < 0) {
		perror("DProbes device open failed");
		return dp_fd;
	}
	return 0;
}

static void dp_device_close(void)
{
	close(dp_fd);
}	
#else

#include <linux/unistd.h>
_syscall3(int, dprobes, unsigned long, cmd, void *, req, void *, result)

static int dp_device_open(void)
{
	return 0;
}

static void dp_device_close(void)
{
	return;
}
#endif

static void dp_remove ()
{
	int retval;

	retval = dprobes(dp_cmd, modname, NULL);
	if (retval < 0)
		perror("dprobes removed failed");
	else  if (dp_cmd & DP_ALL)
		printf("removed dprobes on all modules\n");
	else
		printf("removed dprobes on %s\n", modname);
	return;
}

#define DP_OUTBUFFER_SIZE	4096

static int dp_getvars()
{
	int retval = 0;
	int i = 0;
	unsigned long len;
	unsigned long numvars;
	char * name;
	byte_t * offset = 0;
	byte_t * next_mod = 0;
	struct dp_getvars_in_struct gv_in;
	byte_t * gv_out;

	gv_in.name = modname;
	gv_in.from = index_from ;
	gv_in.to = index_to;
	gv_in.returned = 0;

	if (!(dp_cmd & (DP_GETVARS_LOCAL | DP_GETVARS_GLOBAL)))
		dp_cmd |= DP_GETVARS_LOCAL;

	gv_out = (byte_t *) malloc(DP_OUTBUFFER_SIZE);
	if (!gv_out) {
		retval = -ENOMEM;
		goto err;
	}
	gv_in.allocated = DP_OUTBUFFER_SIZE;
	retval = dprobes(dp_cmd, &gv_in, gv_out);

	if (retval && retval != -ENOMEM) {
		goto done;
	}
	else if (retval == -ENOMEM && gv_in.returned ) {
		gv_out = (byte_t *) malloc(gv_in.returned);
		if (!gv_out) {
			retval = -ENOMEM;
			goto err;
		}
		gv_in.allocated = gv_in.returned;
		retval = dprobes(dp_cmd, &gv_in, gv_out);
		if (retval)
			goto done;
	}

	if (dp_cmd & DP_GETVARS_RESET) {
		printf("Variables are reset to 0.\n");
		goto done;
	}

	if (!gv_in.returned) {
		printf("No information returned\n");
		goto done;
	}

	offset = gv_out;
	if (dp_cmd & DP_GETVARS_GLOBAL) {
		numvars = *((unsigned long*)(offset));
		offset += sizeof(unsigned long);
		if (numvars) {
			printf("Global variables:\n");
			for(i=0; i< numvars; i++) {
				printf ("var %d: %08lx\n", i + gv_in.from,
					*((unsigned long*)(offset)));
				offset += sizeof(unsigned long);
			}
		} else {
			printf("Global variables: none returned.\n");
		}
	}
	if (dp_cmd & DP_GETVARS_LOCAL) {
		if ((offset - gv_out) >= gv_in.returned) {
			printf("Local variables: none returned.\n");
			goto done;
		}
		printf("Local Variables:\n");
		next_mod = offset;
		while (next_mod < (gv_out + gv_in.returned)) {
			len = *((unsigned long*)(offset));
			offset += sizeof(unsigned long);
			numvars = *((unsigned long*)(offset));
			offset += sizeof(unsigned long);
			name = offset + numvars * sizeof(unsigned long);
			printf("Module Name: %s\n", name);
			for(i=0; i< numvars; i++) {
				printf ("var %d: %08lx\n", i + gv_in.from,
					*((unsigned long*)(offset)));
				offset += sizeof(unsigned long);
			}
			next_mod += len;
			offset = next_mod;
		}
	}
done:
	free(gv_out);
err:
	if (retval < 0)
		perror("dprobes getvars failed");
	return retval;

}

static void showprobe(struct dp_outrec_struct *r)
{
	unsigned short probe = r->point.probe & DP_PROBE_MASK;
	if (probe & DP_PROBE_WATCHPOINT)
		printf("  watch");
	else
		printf("  break");
}

static void showtype(struct dp_outrec_struct *r)
{
	unsigned short type = r->point.probe & DP_WATCHTYPE_MASK;

	switch(type) {
		case DP_WATCHTYPE_EXECUTE: printf("    x"); break;
		case DP_WATCHTYPE_WRITE: printf("    w"); break;
		case DP_WATCHTYPE_IO: printf("   io"); break;
		case DP_WATCHTYPE_RDWR: printf("   rw"); break;
	}
}

static void showstatus(struct dp_outrec_struct * r)
{
	unsigned long status = r->status;
	if (!status) {
		printf("invalid-not applied\n");
		return;
	}
	if (status & DP_REC_STATUS_COMPILED)
		printf("compiled");
	if (status & DP_REC_STATUS_ACTIVE) {
		printf(", active\n");
		return;
	}
	if (status & DP_REC_STATUS_MISMATCH) 
		printf(", mismatch (actual 0x%x)", r->point.actual_opcode);
	else if (status & DP_REC_STATUS_INVALID_OFFSET) {
		printf(", invalid %s", (r->point.address_flag & DP_ADDRESS_ABSOLUTE)? "address":"offset");
	}
	else if (status & DP_REC_STATUS_EXCLUDED)
		printf(", excluded region");
	else if (status & DP_REC_STATUS_EXCLUDED_OPCODE)
		printf(", excluded opcode");
	else if (status & DP_REC_STATUS_DISABLED)
		printf(", disabled");
	else if (status & DP_REC_STATUS_REMOVED)
		printf(", removed");
	else if (status & DP_REC_STATUS_DEBUGREG_PATCH_NEEDED)
		printf(", CONFIG_DR_ALLOC needed");
	else if (status & DP_REC_STATUS_DEBUGREG_UNAVAIL)
		printf(", dbreg unavail");
	else if (status & DP_REC_STATUS_WATCHPOINT_LEN_INVALID)
		printf(", invalid watchpoint range");
	printf("\n");
}

static void showrecs(struct dp_outmod_struct *m)
{
	unsigned short num_points = m->pgm.num_points;
	struct dp_outrec_struct * r = m->rec;
	int i = 0;

	if (dp_cmd & DP_QUERY_EXTENDED) {
		printf("major  minor   probe type   addr/off opcode    count status\n");
		for (i = 0; i < num_points; i++, r++) {
			printf("0x%04x", r->point.major);
			printf(" 0x%04x", r->point.minor);
			showprobe(r);
			showtype(r);
			printf(" 0x%08llx", 
				(r->point.address_flag & DP_ADDRESS_ABSOLUTE)? 
				r->point.offset : (r->point.offset - m->base));
			printf("   0x%02x", r->point.opcode);
			printf(" %8ld ", r->count);
			showstatus(r);
		}
	}
	else {
		printf("\tminors: ");
		for (i = 0; i < num_points; i++, r++) {
			printf(" %d", r->point.minor);
		}
		printf("\n");
	}
}

/*
 *
 */
static void showmod(struct dp_outmod_struct * m)
{
	unsigned long * lv;
	int i = 0;

	printf("Module name: %s\n", m->pgm.name);
	if (dp_cmd & DP_QUERY_EXTENDED) {
		printf("\tjmpmax: %d\n", m->pgm.jmpmax);
		printf("\tlogmax: %d\n", m->pgm.logmax);
		lv = m->lv;
		if (lv) {
			printf("\tlocal variables:\n");
			for (i=0; i < m->pgm.num_lv; i++) {
				printf ("\t\tvar %d: 0x%08lx\n", i, *lv);
				lv++;
			}
		}
	}
	showrecs(m);
}

/*
 * caller will have to free the output buffer *res.
 */
static int get_query_buffer(unsigned long qcmd,
	struct dp_query_in_struct *q_in, byte_t **res)
{
	int retval = 0;
	byte_t * result;

	result = (byte_t *) malloc(DP_OUTBUFFER_SIZE);
	if (!result) {
		retval = -ENOMEM;
		goto err;
	}
	q_in->allocated = DP_OUTBUFFER_SIZE;

	retval = dprobes(qcmd, q_in, result);
	if (retval && retval != -ENOMEM && errno != EINVAL)
		goto err_free_res;

	else if (retval == -ENOMEM && q_in->returned ) {
		free(result);
		result = (byte_t *) malloc(q_in->returned);
		if (!result) {
			retval = -ENOMEM;
			goto err;
		}
		q_in->allocated = q_in->returned;
		retval = dprobes(qcmd, q_in, result);
		if (retval)
			goto err_free_res;
	}

	if (!q_in->returned) {
		printf("No information returned.\n");
		goto err_free_res;
	}
	*res = result;
	return retval;

err_free_res:
	free(result);
err:
	*res = NULL;
	if (retval < 0 && errno != EINVAL)
		perror("dprobes query failed");
	else
		retval = 0;
	return retval;
}

static int dp_query()
{
	int retval = 0;
	struct dp_outmod_struct * m;
	struct dp_query_in_struct q_in;
	byte_t * result;

	q_in.name = modname;
	q_in.returned = 0;

	retval = get_query_buffer(dp_cmd, &q_in, &result);
	if (retval < 0 || !result)
		return retval;

	for (m = (struct dp_outmod_struct *)result; m; m = m->next) {
		showmod(m);
	}
	free(result);
	return retval;
}

static int cmdl_system(const char *command)
{
	return system(command);
}

static void dp_preprocessor()
{
	char gcccmd[256] = "\0";
	char *str = gcccmd;
	char tmpfile[40] = "/tmp/rpnXXXXXX";
	int i, retval;
	int fd;

	fd = mkstemp(tmpfile);
	if (fd < 0) {
		perror("tmpfile could not be created\n");
	}
	close(fd);
	prepfile = strdup(tmpfile);

	sprintf(gcccmd, "gcc -E -xc ");
	for(i=0; i < num_defines; i++) {
		str = gcccmd + strlen(gcccmd);
		sprintf(str, "-D%s=%s ", defines_list[i].key, defines_list[i].val);
	}
	str = gcccmd + strlen(gcccmd);
	sprintf(str, "-o %s %s", prepfile, rpnfile);

	retval = cmdl_system(gcccmd);
	if (retval) {
		printf("Pre-processor error. Continuing anyway...\n");
		free(prepfile);
		prepfile = strdup(rpnfile);
	}
	return;
}

static int
cmdl_compare_points(const void *first, const void *second)
{
        unsigned long one = ((const struct dp_point_struct *)first)->offset;
        unsigned long two = ((const struct dp_point_struct *)second)->offset;

        if (one > two)
        	return 1;
	else if (one < two)
	        return -1;
        return 0;
}

static int
copy_point(struct dp_pgm_struct *pgm, struct dp_outrec_struct *old)
{
	struct dp_point_struct *newpt = pgm->point + pgm->num_points;

	*newpt = old->point;
	newpt->rpn_offset += pgm->rpn_length;
	newpt->rpn_end += pgm->rpn_length;
	pgm->num_points++;
	return 0;
}


static int
merge_points(struct dp_outmod_struct *m, struct dp_pgm_struct *pgm)
{
	unsigned short num_new = pgm->num_points;
	unsigned short num_old = m->pgm.num_points;
	struct dp_point_struct *new = pgm->point;
	struct dp_outrec_struct *old = m->rec;
	loff_t old_offset;
	int retval = 0;
	int old_index, new_index;

	/* alloc space for combined point array */
	pgm->point = (struct dp_point_struct *)realloc(pgm->point,
			(num_new + num_old) * sizeof(struct dp_point_struct));
	if (!pgm->point)
		return -ENOMEM;
	new = pgm->point;

	/* copy all old points before the new points */
	for(old_index = 0, new_index = 0; old_index < num_old;
			old_index++, old++) {
		old_offset = old->point.offset;
		if (!(old->point.address_flag & DP_ADDRESS_ABSOLUTE))
			old_offset -= m->base;
		while (old_offset > new->offset) {
			if (new_index == num_new)
				goto out;
			new_index++; new++;
		}

		if (old_offset != new->offset) {
			old->point.offset = old_offset;
			retval = copy_point(pgm, old);
			if (retval < 0)
				return retval;
		}
	}

out:
	/* copy all old points after the new points */
	for ( ; old_index < num_old; old_index++, old++) {
		if (!(old->status & (DP_REC_STATUS_EXCLUDED | 
		DP_REC_STATUS_EXCLUDED_OPCODE | DP_REC_STATUS_MISMATCH))
		&& !(old->point.address_flag & DP_ADDRESS_ABSOLUTE))
			old->point.offset -= m->base;
		retval = copy_point(pgm, old);
		if (retval < 0)
			return retval;
	}

	/* remove unnecessary space in point array */
	pgm->point = (struct dp_point_struct *)realloc(pgm->point,
			(pgm->num_points) * sizeof(struct dp_point_struct));
	if (!pgm->point)
		return -ENOMEM;

	/* copy old rpn_code after the new rpn_code */
	pgm->rpn_code = (byte_t *)realloc(pgm->rpn_code,
		pgm->rpn_length + m->pgm.rpn_length);
	if (!pgm->rpn_code)
		return -ENOMEM;
	memcpy(pgm->rpn_code + pgm->rpn_length, m->pgm.rpn_code,
			m->pgm.rpn_length);
	pgm->rpn_length += m->pgm.rpn_length;

	/* sort the combined point array */
	qsort(pgm->point, pgm->num_points, sizeof(struct dp_point_struct),
                cmdl_compare_points);

	return retval;
}

#define MAX(x, y)  ((x > y) ? x : y)

static void
update_file_header(struct dp_outmod_struct *m, struct dp_pgm_struct *pgm)
{
	pgm->flags = m->pgm.flags;
	pgm->jmpmax = MAX(pgm->jmpmax, m->pgm.jmpmax);
	pgm->logmax = MAX(pgm->logmax, m->pgm.logmax);
	pgm->num_lv = MAX(pgm->num_lv, m->pgm.num_lv);
	pgm->num_gv = MAX(pgm->num_gv, m->pgm.num_gv);
}

static int dp_merge(struct dp_pgm_struct *pgm)
{
	int retval = 0;
	struct dp_outmod_struct * m;
	struct dp_query_in_struct q_in;
	byte_t * result;

	q_in.name = pgm->name;
	q_in.returned = 0;

	retval = get_query_buffer(DP_QUERY, &q_in, &result);
	if ((retval < 0 || !result) && errno != EINVAL)
		return retval;

	m = (struct dp_outmod_struct *) result;
	if (!m)
		goto err_free_res;

	update_file_header(m, pgm);

	retval = merge_points(m, pgm);
	if (retval < 0)
		goto err_free_res;

	retval = dprobes(DP_REMOVE, pgm->name, NULL);
	if (retval < 0)
		goto err_free_res;

	retval = dprobes(DP_INSERT, pgm, NULL);

err_free_res:
	free(result);

	return retval;
}

static int dp_replace(struct dp_pgm_struct *pgm)
{
	int retval = 0;

	retval = dprobes(DP_REMOVE, pgm->name, NULL);
	if (retval < 0)
		return retval;

	return dprobes(DP_INSERT, pgm, NULL);
}

/* if groups are not specified on the cmdline, match all. */
static int match_group(struct dp_point_struct *point)
{
	int i;

	if (!cmdl_group_list.count)
		return 1;

	for (i = 0; i < cmdl_group_list.count; i++) {
		if (point->group == g_index[i])
			return 1;
	}
	return 0;
}

/* if types are not specified on the cmdline, match all. */
static int match_type(struct dp_point_struct *point)
{
	int i;

	if (!cmdl_type_list.count)
		return 1;

	for (i = 0; i < cmdl_type_list.count; i++) {
		if (point->type == t_index[i])
			return 1;
	}
	return 0;
}

static int select_points(struct dp_pgm_struct *pgm)
{
	int i, retval = 0;
	struct dp_point_struct *new_point, *old_point, *n, *o;
	int old_num_points;


	old_point = pgm->point;
	new_point = (struct dp_point_struct *) malloc(pgm->num_points *
			(sizeof(struct dp_point_struct)));
	if (!new_point)
		return -ENOMEM;

	old_num_points = pgm->num_points;
	pgm->num_points = 0;

	for (i = 0, n = new_point, o = old_point; i < old_num_points; i++, o++) {
		if (match_group(o) && match_type(o)) {
			pgm->num_points++;
			*n++ = *o;
		}
	}

	if (!pgm->num_points) {
		free(new_point);
		return -EINVAL;
	}

	new_point = (struct dp_point_struct *)realloc (new_point,
			pgm->num_points	* sizeof(struct dp_point_struct));
	if (!new_point)
		return -ENOMEM;
	pgm->point = new_point;
	free(old_point);
	return retval;
}

static int insert_probes(unsigned long dp_cmd, struct dp_pgm_struct *pgm)
{
	int retval = 0;

	if (cmdl_group_list.count || cmdl_type_list.count) {
		retval = select_points(pgm);
		if (retval)
			return retval;
	}

	dp_cmd &= ~DP_APPLYPPDF;
	retval = dprobes(dp_cmd, pgm, NULL);
	if (retval < 0) {
		if (errno == EEXIST || errno == EINVAL) {
			if (dp_cmd & DP_REPLACE)
				retval = dp_replace(pgm);

			else if (dp_cmd & DP_MERGE)
				retval = dp_merge(pgm);
		}
	}
	return retval;
}

static void dp_insert()
{
	struct dp_pgm_struct pgm;
	int retval;
	int i;

	dp_preprocessor(rpnfile);

	dp_compile(prepfile, symfile, &pgm);

	if (dp_cmd & DP_AUTOSTACKTRACE)
		pgm.autostacktrace = 1;

	pgm.flags |= (dp_cmd & (DP_LOG_MASK | DP_LOG_TARGET_MASK | DP_DONT_VERIFY_OPCODES));
	
	for (i = 0; i < cmdl_group_list.count ; i++)
		g_index[i] = group_index(cmdl_group_list.entry[i].name);
	for (i = 0; i < cmdl_type_list.count ; i++)
		t_index[i] = type_index(cmdl_type_list.entry[i].name);
	
	retval = insert_probes(dp_cmd, &pgm);
	if (retval < 0) {
		if (errno == ENOENT)
			printf("probes cannot be inserted on non-existant module %s\n", pgm.name);
		else if (errno == EEXIST)
			printf("probes already exist on %s\n", pgm.name);
		else
			perror(NULL);
		printf("dprobes insert failed.\n");
	}
	else
		printf("Probes in %s successfully prepared for insertion into %s.\n", 
			rpnfile, pgm.name);

	if (strcmp(prepfile, rpnfile))
		remove(prepfile);

	return ;
}

extern label_table_t group_list;
extern label_table_t type_list;

static int dp_applyppdf()
{
	struct ppdf_hdr_struct hdr;
	struct ppdf_def_struct def;
	struct dp_pgm_struct pgm;
	char ch;
	unsigned long offset = 0;
	int i, fd, retval = 0, name_length = 0;

	/*
	 * for storing the indices of rpngroup and rpntype specified on the
	 * command line.
	 */
	fd = open(ppdffile, O_RDWR, S_IREAD | S_IWRITE);
	if (fd < 0) {
		perror(ppdffile);
		return retval;
	}

	retval = read(fd, &hdr, sizeof(hdr));
	if (retval < 0) {
		perror("ppdf file read error\n");
		return retval;
	}
	if (hdr.sig != PPDF_SIG) {
		printf("file not in ppdf format\n");
		return retval;
	}
	pgm = hdr.pgm;

	group_list.count = hdr.num_groups;
	type_list.count  = hdr.num_types;

        group_list.entry = malloc (group_list.count * sizeof(list_entry_t));
        type_list.entry = malloc (type_list.count * sizeof(list_entry_t));

	offset = hdr.group_offset;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err1;
	for (i = 0; i < hdr.num_groups; i++) {
		retval = lseek(fd, offset, SEEK_SET);
		if (retval < 0)
			goto err1;
		memset(def.name, 0, sizeof(def.name));
		offset += read(fd, def.name, sizeof(def.name));
		group_list.entry[i].name = malloc(sizeof(def.name));
		strncpy(group_list.entry[i].name, def.name, 15);
	}

	offset = hdr.type_offset;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err1;
	for (i = 0; i < hdr.num_types; i++) {
		retval = lseek(fd, offset, SEEK_SET);
		if (retval < 0)
			goto err1;
		memset(def.name, 0, sizeof(def.name));
		offset += read(fd, &def.name, sizeof(def.name));
		type_list.entry[i].name = malloc(sizeof(def.name));
		strncpy(type_list.entry[i].name, def.name, 15);
	}

	for (i = 0; i < cmdl_group_list.count ; i++)
		g_index[i] = group_index(cmdl_group_list.entry[i].name);
	for (i = 0; i < cmdl_type_list.count ; i++)
		t_index[i] = type_index(cmdl_type_list.entry[i].name);

	pgm.point = (struct dp_point_struct *)malloc(pgm.num_points *
					sizeof(struct dp_point_struct));
	if (!pgm.point) {
		retval = -ENOMEM;
		goto err1;
	}

	offset = (unsigned long)hdr.pgm.point;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err2;
	offset += read(fd, pgm.point, pgm.num_points *
				sizeof(struct dp_point_struct));

	pgm.rpn_code = (byte_t *)malloc(pgm.rpn_length * sizeof(pgm.rpn_code));
	if (!pgm.rpn_code) {
		retval = -ENOMEM;
		goto err2;
	}
	offset = (unsigned long)hdr.pgm.rpn_code;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err3;
	offset += read(fd, pgm.rpn_code, pgm.rpn_length);

	offset = (unsigned long)hdr.pgm.name;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err3;
	do {
		retval = read(fd, &ch, 1);
		name_length++;
	} while (ch != '\0');
	name_length++;

	pgm.name = (unsigned char *)malloc(name_length);
	if (!pgm.name) {
		retval = -ENOMEM;
		goto err3;
	}
	offset = (unsigned long)hdr.pgm.name;
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err4;
	retval = read(fd, pgm.name, name_length);

	dp_cmd |= DP_INSERT;
	pgm.flags |= (dp_cmd & (DP_LOG_MASK | DP_LOG_TARGET_MASK));

	retval = insert_probes(dp_cmd, &pgm);
	if (retval < 0 && errno != EEXIST)
		printf("applyppdf failed");
	else
		printf("ppdf applied successfully into %s\n", pgm.name);

err4:	free(pgm.name);
err3:	free(pgm.rpn_code);
err2:	free(pgm.point);
err1:	return retval;
}

static unsigned long write_ppdf(int fd, void * src, int len)
{
	int retval = write(fd, src, len);
	if (retval < len)
		goto err;
	return retval;
err:
	perror("buildppdf failed");
	exit(1);
}


static void dp_ppdf()
{
	struct ppdf_hdr_struct hdr;
	struct ppdf_def_struct def;
	struct dp_pgm_struct pgm;
	unsigned long offset = 0;
	int fd;
	int retval;
	int i;

	dp_preprocessor(rpnfile);

	dp_compile(prepfile, symfile, &pgm);

	pgm.flags |= (dp_cmd & DP_LOG_MASK); //TODO

	hdr.sig = PPDF_SIG;
	hdr.num_groups = group_list.count;
	hdr.num_types = type_list.count;
	hdr.pgm = pgm;

	if (!ppdffile) {
		ppdffile = malloc(strlen(rpnfile) + strlen(PPDF_EXTN) + 1);
		strcpy(ppdffile, rpnfile);
		strcat(ppdffile, PPDF_EXTN);
	}

	fd = open(ppdffile, O_RDWR | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
	if (fd < 0) {
		perror(ppdffile);
		return;
	}

	offset += sizeof(hdr);
	retval = lseek(fd, offset, SEEK_SET);
	if (retval < 0)
		goto err;

	hdr.group_offset = offset;
	for (i = 0; i < hdr.num_groups; i++) {
		memset(def.name, 0, sizeof(def.name));
		strncpy(def.name, group_list.entry[i].name, 15);
		offset += write_ppdf(fd, &def, sizeof(def));
	}

	hdr.type_offset = offset;
	for (i = 0; i < hdr.num_types; i++) {
		memset(def.name, 0, sizeof(def.name));
		strncpy(def.name, type_list.entry[i].name, 15);
		offset += write_ppdf(fd, &def, sizeof(def));
	}

	hdr.pgm.point = (struct dp_point_struct *)offset;
	offset += write_ppdf(fd, pgm.point,
			pgm.num_points * sizeof(struct dp_point_struct));

	hdr.pgm.rpn_code = (byte_t *)offset;
	offset += write_ppdf(fd, pgm.rpn_code, pgm.rpn_length);

	hdr.pgm.name = (char *)offset;
	offset += write_ppdf(fd, pgm.name, strlen(pgm.name)+1); /* null terminated */

	retval = lseek(fd, 0, SEEK_SET);
	if (retval < 0)
		goto err;

	write_ppdf(fd, &hdr, sizeof(hdr));

	retval = remove(prepfile);
	if (retval < 0)
		printf("Could remove pre-processor file %s\n", prepfile);
	return;
err:
	perror("buildppdf failed");
	return;
}

int main(int argc, char * argv[])
{
	int i;
	int len = 0;
	char *pos;

	for (i = 1; i < argc; i++) {
		cmdline_len += strlen(argv[i]) + 1;
	}
	if (!cmdline_len) {
		cmdlerror("No options specified.");
	}

	cmdline = (char *)malloc(cmdline_len);
	if (!cmdline)
		fatal_error("malloc failed\n");
	pos = cmdline;

	for (i = 1; i < argc; i++) {
		len = strlen(argv[i]);
		strncpy(pos, argv[i], len);
		pos += len;
		*pos++ = ' ';
	}

	*(--pos) = '\0';

	cmdline_ptr = cmdline;
	cmdline_end = cmdline + cmdline_len;

	max_defines = 32;
	defines_list = (struct defines *)malloc(sizeof(struct defines) * max_defines);
	if (!defines_list)
		fatal_error("malloc failed\n");

	cmdlparse();

	if (dp_device_open() < 0) 
		return -1;

       	switch (dp_cmd & DP_CMD_MASK) {
        case DP_INSERT:
                dp_insert();
                break;
        case DP_REMOVE:
                dp_remove();
                break;
        case DP_GETVARS:
                dp_getvars();
                break;
        case DP_QUERY:
                dp_query();
                break;
        case DP_VERSION:
		display_version();
                break;
        case DP_HELP:
		display_help();
                break;
        case DP_BUILDPPDF:
                dp_ppdf();
                break;
	case DP_APPLYPPDF:
		dp_applyppdf();
		break;
        default:
		printf("dprobes: missing/incorrect arguments\n");
		printf("Try `dprobes --help' for more information\n");
		break;
	}
	dp_device_close();

	return 0;
}


unsigned short dp_mod_type;

struct dp_pgm_struct pgm;
struct dp_point_struct *point;
unsigned int cur_point;
static unsigned int max_point;

unsigned char * rpn_code;
unsigned int rpn_ip = 0;
unsigned int max_rpn;

char * filename;
static char * symfile_from_cmd;

unsigned char logonfault = 0;
int probe_by_symbol = 0;

static label_table_t jmp_target = {"jump targets", NUM_JMPS, 0, NULL};
static label_table_t call_target = {"call targets", NUM_CALLS, 0, NULL};
static label_table_t jmp_list = {"jmp list", NUM_JMPS, 0, NULL};
static label_table_t call_list = {"call list", NUM_CALLS, 0, NULL};
label_table_t group_list = {"group list", NUM_GROUPS, 0, NULL};
label_table_t type_list = {"type list", NUM_TYPES, 0, NULL};

/* 
 * The following function needs to be fixed for probes on User apps and Kernel Modules  
 */
unsigned long fix_offset(unsigned long addr)
{
	if (pgm.flags & DP_MODTYPE_KERNEL)
		addr += PAGE_OFFSET;
	return addr;
}		

void fix_name()
{
	if (pgm.flags & DP_MODTYPE_KMOD) {
		char * name;
		char * p;
		int len;

		if ((p = strrchr(pgm.name, '/')) != NULL)
			p++;
		else 
			p = pgm.name;	
		len = strlen(p);
		if (len > 2 && strcmp(p + len - 2, ".o") == 0)
			len -= 2;
		else if (len > 4 && strcmp(p + len - 4, ".mod") == 0)
			len -= 4;
		name = malloc(len +1);
		if (!name) 
			fatal_error("malloc failed\n");
		memcpy(name, p, len);
		name[len] = '\0';
		free(pgm.name);
		pgm.name = name;
	}
}

void end_file_header()
{
	if (symfile_from_cmd)
		init_syms(symfile_from_cmd);
	else if(!(pgm.flags & DP_MODTYPE_KERNEL)) /* --sym required for kernel symbol support. */ 
		init_syms(pgm.name);

	if (pgm.flags & DP_MODTYPE_KMOD) 
		pgm.align = get_kmod_align(); 
	fix_name();

	/* set default values if not specified in the rpn */
	if (!(pgm.flags & DP_MODTYPE_MASK)) pgm.flags |= DP_MODTYPE_USER; 
}

void start_probe(unsigned long offset, unsigned long len, unsigned long flag)
{
	/* initialize this point */
	point[cur_point].major = pgm.major;
	point[cur_point].rpn_offset = rpn_ip;
	point[cur_point].maxhits = DEFAULT_MAXHITS;
	point[cur_point].passcount = 0;
	point[cur_point].group = 0;
	point[cur_point].type = 0;
	if (flag & DP_ADDRESS_ABSOLUTE && pgm.flags & DP_MODTYPE_USER) 
		rpnerror("Absolute address can't be specified for probes in user modules.");
	point[cur_point].address_flag = flag;
	point[cur_point].offset = offset;
	point[cur_point].len = len;
	point[cur_point].probe = (DP_PROBE_BREAKPOINT | DP_WATCHTYPE_EXECUTE);
	point[cur_point].logonfault = logonfault;
	point[cur_point].ex_mask = DEFAULT_EX_MASK;	
}

void set_lvars(unsigned long n)
{
	char buf[80];
	if (n > DP_MAX_LVARS) {
		sprintf(buf, "num_lv can not be > %d", DP_MAX_LVARS);
		rpnerror(buf);
	} else
		pgm.num_lv = n;
}

void set_gvars(unsigned long n)
{
	char buf[80];
	if (n > DP_MAX_GVARS) {
		sprintf(buf, "num_gv can not be > %d", DP_MAX_GVARS);
		rpnerror(buf);
	} else
		pgm.num_gv = n;
}

void fix_address(void)
{
	if (point[cur_point].address_flag & DP_ADDRESS_ABSOLUTE)
		if (point[cur_point].offset < PAGE_OFFSET)
			point[cur_point].offset += PAGE_OFFSET;
}

void set_probe_type(void)
{
		if (pgm.flags & (DP_MODTYPE_USER | DP_MODTYPE_KMOD))
			rpnerror("watchpoint type of probepoints can be specified only in kernel mode rpn files");
		point[cur_point].probe &= ~DP_PROBE_BREAKPOINT;
		point[cur_point].probe |= DP_PROBE_WATCHPOINT;
}

void end_probe()
{
	/* ensure that opcode is specified */
	if ((point[cur_point].probe & DP_PROBE_BREAKPOINT) &&
		!point[cur_point].opcode &&
		(!(dp_cmd & DP_DONT_VERIFY_OPCODES) || !probe_by_symbol)) {
		rpnerror("Opcode is not specified for this probe point.");
	}
	probe_by_symbol = 0;
	
	point[cur_point].rpn_end = rpn_ip;

	if (++cur_point >= max_point) {
		max_point += NUM_POINTS;
		point = (struct dp_point_struct *)realloc((void *)point,
			(max_point * sizeof(*point)));
		if (!point) {
			fatal_error("realloc failed\n");
		}
	}
}

void ins_size(int s)
{
	if (rpn_ip + s >= max_rpn) {
		max_rpn += CODE_BUF_SIZE;
		rpn_code = (char *)realloc((void *)rpn_code, max_rpn);
		if (!rpn_code) {
			fatal_error("realloc failed\n");
		}
	}
}

void add_to_list (label_table_t *list, char *name, unsigned short off)
{
	char buff[80];
	int i = list->count;

	if (i >= list->max) {
		sprintf(buff, "Overflow in %s", list->type_name);
		rpnerror(buff);
	}
 	list->entry[i].name = name;
 	list->entry[i].offset = off;
 	list->entry[i].lineno = rpnlineno;
 	list->count++;
}

void add_call(char * name)
{
        add_to_list(&call_list, name, rpn_ip);
}

void add_jmp(char * name)
{
	add_to_list(&jmp_list, name, rpn_ip);
}

list_entry_t * lookup(label_table_t * list, const char * name)
{
	int i;
	for (i = 0; i < list->count; i++) {
		if (!strcasecmp(list->entry[i].name, name)) {
			return &list->entry[i];
		}
	}
	return NULL;

}

void add_call_target(char * name)
{
	char buff[80];
	list_entry_t * dup;

	dup = lookup(&call_target, name);
	if (dup) {
		sprintf(buff, 
			"procedure %s is duplicate. Original is on line %d",
			name, dup->lineno);
		rpnerror(buff);
	}
	add_to_list(&call_target, name, rpn_ip);
}

void add_jmp_target(char * name)
{
	char buff[80];
	list_entry_t * dup;

	dup = lookup(&jmp_target, name);
	if (dup) {
		sprintf(buff, "label %s is duplicate. Original is on line %d",
			name, dup->lineno);
		rpnerror(buff);
	}
	add_to_list(&jmp_target, name, rpn_ip);
}

void add_group(char * name)
{
	char buff[32];
	list_entry_t * dup;

	if (strlen(name) > 15) {
		printf("Warning: group %s truncated to 15 characters.\n", name);
		name[15] = '\0';
	}

	dup = lookup(&group_list, name);
	if (dup) {
		sprintf(buff, "group %s is duplicate. Original is on line %d",
			name, dup->lineno);
		rpnerror(buff);
	}
	add_to_list(&group_list, name, 0);
}

void add_type(char * name)
{
	char buff[32];
	list_entry_t * dup;

	if (strlen(name) > 15) {
		printf("Warning: type %s truncated to 15 characters.\n", name);
		name[15] = '\0';
	}

	dup = lookup(&type_list, name);
	if (dup) {
		sprintf(buff, "type %s is duplicate. Original is on line %d",
			name, dup->lineno);
		rpnerror(buff);
	}
	add_to_list(&type_list, name, 0);
}
int lookup_index(label_table_t * list, const char * name)
{
	int i;
	for (i = 0; i < list->count; i++) {
		if (!strcasecmp(list->entry[i].name, name)) {
			return i;
		}
	}
	return -1;
}

int group_index(char * name)
{
	char buff[32];
	int i;
	i = lookup_index(&group_list, name);
	
	if (i < 0) {
		sprintf(buff, "group %s is undefined", name);
		rpnerror(buff);
	}
	return i;
}

int type_index(char * name)
{
	char buff[32];
	int i;
	i = lookup_index(&type_list, name);
	
	if (i < 0) {
		sprintf(buff, "type %s is undefined", name);
		rpnerror(buff);
	}
	return i;
}

void fixup(label_table_t *list, label_table_t *target)
{
	int i;
	char buff[80];
	list_entry_t * e = list->entry;
	list_entry_t * entry;

	for (i=0; i < list->count; i++, e++) {
		entry = lookup (target, e->name);
		if (!entry) {
			sprintf(buff, "jmp target not found for %s on line %d",
				e->name, e->lineno);
			rpnerror(buff);
		}
		*(long *)(&rpn_code[e->offset-(sizeof(long))]) = 
					(long)(entry->offset - e->offset);
		free(e->name);
		e->name = NULL;
	}

	list->count = 0;

	e = target->entry;
	for (i=0; i < target->count; i++, e++) {
		free(e->name);
		e->name = NULL;
	}
	target->count = 0;
}

void fixup_jmps()
{
 	fixup(&jmp_list, &jmp_target);
}

void fixup_calls()
{
 	fixup(&call_list, &call_target);
}
int rpn_compare_points(const void *first, const void *second)
{
	unsigned long one = ((const struct dp_point_struct *)first)->offset;
	unsigned long two = ((const struct dp_point_struct *)second)->offset;
	
	if (one > two)
		return 1;
	else if (one < two)
		return -1;
	rpnerror("You cannot place two probes at the same location.\n");
	return 0;
}

int dp_compile(char *prepfile, char *symfile, struct dp_pgm_struct *p)
{
	extern FILE *rpnin;

	rpnin = fopen(prepfile, "r");
	if (!rpnin) {
		printf("%s fopen failed\n", prepfile); 
		exit(1);
	}
	filename = prepfile;

        point = (struct dp_point_struct *)malloc(NUM_POINTS * sizeof(*point));
        rpn_code = (char *)malloc(CODE_BUF_SIZE);
        call_list.entry = malloc (call_list.max * sizeof(list_entry_t));
        jmp_list.entry = malloc (jmp_list.max * sizeof(list_entry_t));
        call_target.entry = malloc (call_target.max * sizeof(list_entry_t));
        jmp_target.entry = malloc (jmp_target.max * sizeof(list_entry_t));
        group_list.entry = malloc (group_list.max * sizeof(list_entry_t));
        type_list.entry = malloc (type_list.max * sizeof(list_entry_t));

        if (!point || !rpn_code ||
            !call_list.entry || !jmp_list.entry ||
            !call_target.entry || !jmp_target.entry ||
            !group_list.entry || !type_list.entry) {
		printf("malloc failed\n");
		exit(1);
        }

	/* 
	 * initialize default values of pgm struct. These will be overridden 
	 * if they are specified in the file header. 
	 */
	pgm.logmax = DEFAULT_LOGMAX;
	pgm.ex_logmax = DEFAULT_EX_LOG_MAX;
	pgm.jmpmax = DEFAULT_JMPMAX;

        cur_point = 0;
        max_point = NUM_POINTS;
        rpn_ip = 0;
        max_rpn = CODE_BUF_SIZE;

	symfile_from_cmd  = symfile;
	
	rpnparse();

	fixup_calls();

	pgm.rpn_length = rpn_ip;	
	pgm.rpn_code = rpn_code;

	pgm.num_points = cur_point;
	qsort(point, pgm.num_points, sizeof(*point), rpn_compare_points);
	pgm.point = point;

	*p = pgm;

	cleanup_syms();
}
